@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2001-2020 the xine project
+ * Copyright (C) 2001-2021 the xine project
  *
  * This file is part of xine, a free video player.
  *
@@ -74,7 +74,7 @@ typedef struct ff_audio_decoder_s {
   int               size;
 
   AVCodecContext    *context;
-  AVCodec           *codec;
+  const AVCodec     *codec;
 
   char              *decode_buffer;
   int               decoder_ok;
@@ -83,6 +83,9 @@ typedef struct ff_audio_decoder_s {
 #if XFF_AUDIO > 3
   AVFrame          *av_frame;
 #endif
+#if XFF_AUDIO > 2
+  XFF_PACKET_DECL (avpkt);
+#endif
 
   /* AAC ADTS */
   uint32_t          buftype;
@@ -298,6 +301,7 @@ static void ff_audio_init_codec(ff_audio_decoder_t *this, unsigned int codec_typ
 #if XFF_PARSE > 1
   if (codec_type == BUF_AUDIO_AAC_LATM ||
       codec_type == BUF_AUDIO_EAC3 ||
+      codec_type == BUF_AUDIO_A52 ||
       codec_type == BUF_AUDIO_MPEG) {
 
     xprintf (this->stream->xine, XINE_VERBOSITY_DEBUG,
@@ -690,11 +694,9 @@ static int ff_audio_decode (ff_audio_decoder_t *this,
   else parsed = 0;
 
 #if XFF_AUDIO > 2
-  AVPacket avpkt;
-  av_init_packet (&avpkt);
-  avpkt.data = buf;
-  avpkt.size = size;
-  avpkt.flags = AV_PKT_FLAG_KEY;
+  this->avpkt->data = buf;
+  this->avpkt->size = size;
+  this->avpkt->flags = AV_PKT_FLAG_KEY;
 #  if XFF_AUDIO > 3
   int got_frame;
   float gain = this->class->gain;
@@ -702,7 +704,7 @@ static int ff_audio_decode (ff_audio_decoder_t *this,
     this->av_frame = XFF_ALLOC_FRAME ();
 #   if XFF_AUDIO == 5
   {
-    int err = avcodec_send_packet (this->context, &avpkt);
+    int err = avcodec_send_packet (this->context, this->avpkt);
     /* xprintf (this->stream->xine, XINE_VERBOSITY_DEBUG, "ff_audio_dec: send (%d) = %d.\n", (int)size, err); */
     /* multiple frames per packet */
     consumed = (err >= 0) ? size : ((err == AVERROR (EAGAIN)) ? 0 : err);
@@ -712,7 +714,7 @@ static int ff_audio_decode (ff_audio_decoder_t *this,
     got_frame = (err == 0);
   }
 #   else
-  consumed = avcodec_decode_audio4 (this->context, this->av_frame, &got_frame, &avpkt);
+  consumed = avcodec_decode_audio4 (this->context, this->av_frame, &got_frame, this->avpkt);
 #   endif
   if ((consumed >= 0) && got_frame) {
     /* setup may have altered while decoding */
@@ -970,7 +972,7 @@ static int ff_audio_decode (ff_audio_decoder_t *this,
     }
   } else *decode_buffer_size = 0;
 #  else
-  consumed = avcodec_decode_audio3 (this->context, decode_buffer, decode_buffer_size, &avpkt);
+  consumed = avcodec_decode_audio3 (this->context, decode_buffer, decode_buffer_size, this->avpkt);
   ff_map_channels (this);
 #  endif
 #else
@@ -1286,6 +1288,8 @@ static void ff_audio_dispose (audio_decoder_t *this_gen) {
   this->context->extradata_size = 0;
   XFF_FREE_CONTEXT (this->context);
 
+  XFF_PACKET_UNREF (this->avpkt);
+
   free (this_gen);
 }
 
@@ -1313,6 +1317,10 @@ static audio_decoder_t *ff_audio_open_plugin (audio_decoder_class_t *class_gen,
 # endif
 #endif
 
+# if XFF_AUDIO > 2
+  XFF_PACKET_NEW (this->avpkt);
+#endif
+
   this->class  = (ff_audio_class_t *)class_gen;
   this->stream = stream;
 
