@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2001-2013 the xine project
+ * Copyright (C) 2001-2018 the xine project
  *
  * This file is part of xine, a free video player.
  *
@@ -52,22 +52,15 @@ typedef struct {
   demux_plugin_t       demux_plugin;
 
   xine_stream_t       *stream;
-  fifo_buffer_t       *video_fifo;
   fifo_buffer_t       *audio_fifo;
   input_plugin_t      *input;
   int                  status;
 
-  off_t                data_size;
-
   int                  seek_flag;  /* this is set when a seek just occurred */
 } demux_aac_t;
 
-typedef struct {
-  demux_class_t     demux_class;
-} demux_aac_class_t;
 
-
-static int open_aac_file(demux_aac_t *this) {
+static int probe_aac_file(xine_stream_t *stream, input_plugin_t *input) {
   int i;
   uint8_t peak[MAX_PREVIEW_SIZE];
   uint32_t signature;
@@ -77,17 +70,18 @@ static int open_aac_file(demux_aac_t *this) {
 
   _x_assert(MAX_PREVIEW_SIZE > 10);
 
-  if (_x_demux_read_header(this->input, &signature, 4) != 4)
+  if (_x_demux_read_header(input, &signature, 4) != 4)
       return 0;
 
   /* Check if there's an ID3v2 tag at the start */
   if ( id3v2_istag(signature) ) {
-    this->input->seek(this->input, 4, SEEK_SET);
+    if (input->seek(input, 4, SEEK_SET) != 4)
+      return 0;
 
-    id3v2_parse_tag(this->input, this->stream, signature);
+    id3v2_parse_tag(input, stream, signature);
   }
 
-  if ( this->input->read(this->input, &signature, 4) != 4 )
+  if (input->read(input, &signature, 4) != 4)
     return 0;
 
   /* Check for an ADIF header - should be at the start of the file */
@@ -97,14 +91,16 @@ static int open_aac_file(demux_aac_t *this) {
   }
 
   /* Look for an ADTS header - might not be at the start of the file */
-  if ( this->input->get_capabilities(this->input) & INPUT_CAP_SEEKABLE ) {
+  if (input->get_capabilities(input) & INPUT_CAP_SEEKABLE) {
     lprintf("Getting a buffer of size %u\n", MAX_PREVIEW_SIZE);
 
-    if ( this->input->read(this->input, peak, MAX_PREVIEW_SIZE) != MAX_PREVIEW_SIZE )
+    if (input->read(input, peak, MAX_PREVIEW_SIZE) != MAX_PREVIEW_SIZE )
       return 0;
-    this->input->seek(this->input, 0, SEEK_SET);
-  } else if (_x_demux_read_header(this->input, peak, MAX_PREVIEW_SIZE) !=
-	   MAX_PREVIEW_SIZE)
+    if (input->seek(input, 0, SEEK_SET) != 0)
+      return 0;
+
+  } else if (_x_demux_read_header(input, peak, MAX_PREVIEW_SIZE) !=
+             MAX_PREVIEW_SIZE)
     return 0;
 
   for (i=0; i<MAX_PREVIEW_SIZE; i++) {
@@ -137,10 +133,8 @@ static int open_aac_file(demux_aac_t *this) {
     {
       lprintf("found second ADTS header\n");
 
-      _x_stream_info_set(this->stream, XINE_STREAM_INFO_HAS_VIDEO, 0);
-      _x_stream_info_set(this->stream, XINE_STREAM_INFO_HAS_AUDIO, 1);
-
-      this->input->seek(this->input, data_start+id3size, SEEK_SET);
+      if (input->seek(input, data_start+id3size, SEEK_SET) < 0)
+        return 0;
       return 1;
     }
   }
@@ -191,7 +185,6 @@ static void demux_aac_send_headers(demux_plugin_t *this_gen) {
   demux_aac_t *this = (demux_aac_t *) this_gen;
   buf_element_t *buf;
 
-  this->video_fifo  = this->stream->video_fifo;
   this->audio_fifo  = this->stream->audio_fifo;
 
   this->status = DEMUX_OK;
@@ -214,6 +207,8 @@ static int demux_aac_seek (demux_plugin_t *this_gen,
                            off_t start_pos, int start_time, int playing) {
   demux_aac_t *this = (demux_aac_t *) this_gen;
 
+  (void)start_pos;
+  (void)start_time;
   /* if thread is not running, initialize demuxer */
   if( !playing ) {
 
@@ -235,15 +230,20 @@ static int demux_aac_get_status (demux_plugin_t *this_gen) {
 static int demux_aac_get_stream_length (demux_plugin_t *this_gen) {
 //  demux_aac_t *this = (demux_aac_t *) this_gen;
 
+  (void)this_gen;
   return 0;
 }
 
 static uint32_t demux_aac_get_capabilities(demux_plugin_t *this_gen) {
+  (void)this_gen;
   return DEMUX_CAP_NOCAP;
 }
 
 static int demux_aac_get_optional_data(demux_plugin_t *this_gen,
                                        void *data, int data_type) {
+  (void)this_gen;
+  (void)data;
+  (void)data_type;
   return DEMUX_OPTIONAL_UNSUPPORTED;
 }
 
@@ -252,7 +252,21 @@ static demux_plugin_t *open_plugin (demux_class_t *class_gen, xine_stream_t *str
 
   demux_aac_t    *this;
 
-  this         = calloc(1, sizeof(demux_aac_t));
+  switch (stream->content_detection_method) {
+    case METHOD_BY_MRL:
+    case METHOD_BY_CONTENT:
+    case METHOD_EXPLICIT:
+      if (!probe_aac_file(stream, input))
+        return NULL;
+      break;
+    default:
+      return NULL;
+  }
+
+  this = calloc(1, sizeof(demux_aac_t));
+  if (!this)
+    return NULL;
+
   this->stream = stream;
   this->input  = input;
 
@@ -267,37 +281,27 @@ static demux_plugin_t *open_plugin (demux_class_t *class_gen, xine_stream_t *str
   this->demux_plugin.demux_class       = class_gen;
 
   this->status = DEMUX_FINISHED;
-  switch (stream->content_detection_method) {
 
-  case METHOD_BY_MRL:
-  case METHOD_BY_CONTENT:
-  case METHOD_EXPLICIT:
-    if (!open_aac_file(this)) {
-      free (this);
-      return NULL;
-    }
-    break;
-
-  default:
-    free (this);
-    return NULL;
-  }
+  _x_stream_info_set(stream, XINE_STREAM_INFO_HAS_VIDEO, 0);
+  _x_stream_info_set(stream, XINE_STREAM_INFO_HAS_AUDIO, 1);
 
   return &this->demux_plugin;
 }
 
-void *demux_aac_init_plugin (xine_t *xine, void *data) {
-  demux_aac_class_t     *this;
+void *demux_aac_init_plugin (xine_t *xine, const void *data) {
 
-  this = calloc(1, sizeof(demux_aac_class_t));
+  (void)xine;
+  (void)data;
 
-  this->demux_class.open_plugin     = open_plugin;
-  this->demux_class.description     = N_("ADIF/ADTS AAC demux plugin");
-  this->demux_class.identifier      = "AAC";
-  this->demux_class.mimetypes       = NULL;
-  this->demux_class.extensions      = "aac";
-  this->demux_class.dispose         = default_demux_class_dispose;
+  static const demux_class_t demux_aac_class = {
+    .open_plugin     = open_plugin,
+    .description     = N_("ADIF/ADTS AAC demux plugin"),
+    .identifier      = "AAC",
+    .mimetypes       = NULL,
+    .extensions      = "aac",
+    .dispose         = NULL,
+  };
 
-  return this;
+  return (void*)&demux_aac_class;
 }
 
