@@ -742,6 +742,7 @@ static int read_flv_packet(demux_flv_t *this, int preview) {
             /* better: +/- 16 frames, but we cannot trust header framerate */
             if ((ptsoffs < -1000) || (ptsoffs > 1000))
               ptsoffs = 0;
+            ptsoffs *= 90;
           break;
           case VF_VP6:
             buf_type = BUF_VIDEO_VP6F;
@@ -847,8 +848,6 @@ static int read_flv_packet(demux_flv_t *this, int preview) {
     /* send frame contents */
     buf_pts = (int64_t)pts * 90;
     check_newpts (this, buf_pts, (tag_type == FLV_TAG_TYPE_VIDEO));
-    pts += ptsoffs;
-    buf_pts += ptsoffs * 90;
     size = this->input->get_length (this->input);
     if (size > 0) {
       this->size = size;
@@ -863,7 +862,7 @@ static int read_flv_packet(demux_flv_t *this, int preview) {
     while (remaining_bytes) {
       buf       = fifo->buffer_pool_size_alloc (fifo, remaining_bytes);
       buf->type = buf_type;
-      buf->pts  = buf_pts;
+      buf->pts  = buf_pts + ptsoffs;
       buf->extra_info->input_time = pts;
       if (size > 0)
         buf->extra_info->input_normpos = normpos;
