@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2000-2007 the xine project and Fredrik Noring
+ * Copyright (C) 2000-2021 the xine project and Fredrik Noring
  *
  * This file is part of xine, a free video player.
  *
@@ -65,11 +65,7 @@
 #include <pthread.h>
 #include <netinet/in.h>
 
-#ifdef HAVE_FFMPEG_AVUTIL_H
-#  include <mem.h>
-#else
-#  include <libavutil/mem.h>
-#endif
+
 
 #include <linux/fb.h>
 #include <linux/kd.h>
@@ -106,8 +102,6 @@ typedef struct fb_frame_s
   uint8_t*           video_mem;            /* mmapped video memory */
   uint8_t*           data;
   int                yoffset;
-
-  struct fb_driver_s *this;
 } fb_frame_t;
 
 typedef struct fb_driver_s
@@ -152,18 +146,18 @@ typedef struct fb_driver_s
 typedef struct
 {
   video_driver_class_t driver_class;
-  config_values_t     *config;
   xine_t              *xine;
 } fb_class_t;
 
 static uint32_t fb_get_capabilities(vo_driver_t *this_gen)
 {
+  (void)this_gen;
   return VO_CAP_YV12 | VO_CAP_YUY2 | VO_CAP_BRIGHTNESS | VO_CAP_CONTRAST | VO_CAP_SATURATION;
 }
 
 static void fb_frame_proc_slice(vo_frame_t *vo_img, uint8_t **src)
 {
-  fb_frame_t *frame = (fb_frame_t *)vo_img ;
+  fb_frame_t *frame = xine_container_of(vo_img, fb_frame_t, vo_frame);
 
   vo_img->proc_called = 1;
 
@@ -186,7 +180,7 @@ static void fb_frame_proc_slice(vo_frame_t *vo_img, uint8_t **src)
 
 static void fb_frame_field(vo_frame_t *vo_img, int which_field)
 {
-  fb_frame_t *frame = (fb_frame_t *)vo_img ;
+  fb_frame_t *frame = xine_container_of(vo_img, fb_frame_t, vo_frame);
 
   switch(which_field)
   {
@@ -208,16 +202,19 @@ static void fb_frame_field(vo_frame_t *vo_img, int which_field)
 
 static void fb_frame_dispose(vo_frame_t *vo_img)
 {
-  fb_frame_t *frame = (fb_frame_t *)vo_img;
+  fb_frame_t *frame = xine_container_of(vo_img, fb_frame_t, vo_frame);
+  fb_driver_t *this = xine_container_of(vo_img->driver, fb_driver_t, vo_driver);
 
-  if(!frame->this->use_zero_copy)
+  frame->yuv2rgb->dispose (frame->yuv2rgb);
+
+  if(!this->use_zero_copy)
      free(frame->data);
   free(frame);
 }
 
 static vo_frame_t *fb_alloc_frame(vo_driver_t *this_gen)
 {
-  fb_driver_t *this = (fb_driver_t *)this_gen;
+  fb_driver_t *this = xine_container_of(this_gen, fb_driver_t, vo_driver);
   fb_frame_t *frame;
 
   if(this->use_zero_copy &&
@@ -228,6 +225,13 @@ static vo_frame_t *fb_alloc_frame(vo_driver_t *this_gen)
   if(!frame)
     return NULL;
 
+  /* colorspace converter for this frame */
+  frame->yuv2rgb = this->yuv2rgb_factory->create_converter(this->yuv2rgb_factory);
+  if (!frame->yuv2rgb) {
+    free(frame);
+    return NULL;
+  }
+
   memcpy(&frame->sc, &this->sc, sizeof(vo_scale_t));
 
   pthread_mutex_init(&frame->vo_frame.mutex, NULL);
@@ -239,12 +243,6 @@ static vo_frame_t *fb_alloc_frame(vo_driver_t *this_gen)
   frame->vo_frame.dispose    = fb_frame_dispose;
   frame->vo_frame.driver     = this_gen;
 
-  frame->this = this;
-
-  /* colorspace converter for this frame */
-  frame->yuv2rgb =
-    this->yuv2rgb_factory->create_converter(this->yuv2rgb_factory);
-
   if(this->use_zero_copy)
   {
     frame->yoffset = this->used_num_buffers * this->fb_var.yres;
@@ -260,16 +258,19 @@ static vo_frame_t *fb_alloc_frame(vo_driver_t *this_gen)
 
   this->used_num_buffers++;
 
-  return (vo_frame_t *)frame;
+  return &frame->vo_frame;
 }
 
 static void fb_compute_ideal_size(fb_driver_t *this, fb_frame_t *frame)
 {
+  (void)this;
+
   _x_vo_scale_compute_ideal_size(&frame->sc);
 }
 
 static void fb_compute_rgb_size(fb_driver_t *this, fb_frame_t *frame)
 {
+  (void)this;
   _x_vo_scale_compute_output_size(&frame->sc);
 
   /* avoid problems in yuv2rgb */
@@ -327,9 +328,9 @@ static void setup_colorspace_converter(fb_frame_t *frame, int flags)
 static void frame_reallocate(fb_driver_t *this, fb_frame_t *frame,
 			     uint32_t width, uint32_t height, int format)
 {
-  av_freep(&frame->vo_frame.base[0]);
-  av_freep(&frame->vo_frame.base[1]);
-  av_freep(&frame->vo_frame.base[2]);
+  xine_freep_aligned(&frame->vo_frame.base[0]);
+  xine_freep_aligned(&frame->vo_frame.base[1]);
+  xine_freep_aligned(&frame->vo_frame.base[2]);
 
   if(this->use_zero_copy)
   {
@@ -351,15 +352,15 @@ static void frame_reallocate(fb_driver_t *this, fb_frame_t *frame,
       frame->vo_frame.pitches[1] = 8*((width + 15) / 16);
       frame->vo_frame.pitches[2] = 8*((width + 15) / 16);
 
-      frame->vo_frame.base[0] = av_mallocz(frame->vo_frame.pitches[0] * height);
-      frame->vo_frame.base[1] = av_mallocz(frame->vo_frame.pitches[1] * ((height+1)/2));
-      frame->vo_frame.base[2] = av_mallocz(frame->vo_frame.pitches[2] * ((height+1)/2));
+      frame->vo_frame.base[0] = xine_mallocz_aligned(frame->vo_frame.pitches[0] * height);
+      frame->vo_frame.base[1] = xine_mallocz_aligned(frame->vo_frame.pitches[1] * ((height+1)/2));
+      frame->vo_frame.base[2] = xine_mallocz_aligned(frame->vo_frame.pitches[2] * ((height+1)/2));
   }
   else
   {
     frame->vo_frame.pitches[0] = 8 * ((width + 3) / 4);
 
-    frame->vo_frame.base[0] = av_mallocz(frame->vo_frame.pitches[0] * height);
+    frame->vo_frame.base[0] = xine_mallocz_aligned(frame->vo_frame.pitches[0] * height);
   }
 }
 
@@ -368,17 +369,17 @@ static void fb_update_frame_format(vo_driver_t *this_gen,
 				   uint32_t width, uint32_t height,
 				   double ratio, int format, int flags)
 {
-  fb_driver_t *this = (fb_driver_t *)this_gen;
-  fb_frame_t *frame = (fb_frame_t *)frame_gen;
+  fb_driver_t *this  = xine_container_of(this_gen,  fb_driver_t, vo_driver);
+  fb_frame_t  *frame = xine_container_of(frame_gen, fb_frame_t,  vo_frame);
 
   flags &= VO_BOTH_FIELDS;
 
   /* Find out if we need to adapt this frame. */
-  if (width != frame->sc.delivered_width    ||
-      height != frame->sc.delivered_height  ||
-      ratio != frame->sc.delivered_ratio    ||
-      flags != frame->flags                 ||
-      format != frame->format               ||
+  if ((int)width  != frame->sc.delivered_width  ||
+      (int)height != frame->sc.delivered_height ||
+      ratio       != frame->sc.delivered_ratio  ||
+      flags       != frame->flags               ||
+      format      != frame->format              ||
       this->sc.user_ratio != frame->sc.user_ratio)
   {
     lprintf("frame format (from decoder) has changed => adapt\n");
@@ -410,39 +411,24 @@ static void fb_update_frame_format(vo_driver_t *this_gen,
 static void fb_overlay_clut_yuv2rgb(fb_driver_t *this,
 				    vo_overlay_t *overlay, fb_frame_t *frame)
 {
-  size_t i;
-  clut_t* clut = (clut_t*)overlay->color;
+  int i;
+  uint32_t *rgb;
 
-  if(!overlay->rgb_clut)
-  {
-    for(i = 0;
-	i < sizeof(overlay->color)/sizeof(overlay->color[0]);
-	i++)
-    {
-      *((uint32_t *)&clut[i]) =
-	frame->yuv2rgb->
-	yuv2rgb_single_pixel_fun(frame->yuv2rgb,
-				 clut[i].y,
-				 clut[i].cb,
-				 clut[i].cr);
+  (void)this;
+  if (!overlay->rgb_clut) {
+    rgb = overlay->color;
+    for (i = sizeof (overlay->color) / sizeof (overlay->color[0]); i > 0; i--) {
+      clut_t *yuv = (clut_t *)rgb;
+      *rgb++ = frame->yuv2rgb->yuv2rgb_single_pixel_fun (frame->yuv2rgb, yuv->y, yuv->cb, yuv->cr);
     }
     overlay->rgb_clut++;
   }
 
-  if(!overlay->hili_rgb_clut)
-  {
-    clut = (clut_t*) overlay->hili_color;
-
-    for(i = 0;
-	i < sizeof(overlay->color)/sizeof(overlay->color[0]);
-	i++)
-    {
-      *((uint32_t *)&clut[i]) =
-	frame->yuv2rgb->
-	yuv2rgb_single_pixel_fun(frame->yuv2rgb,
-				 clut[i].y,
-				 clut[i].cb,
-				 clut[i].cr);
+  if (!overlay->hili_rgb_clut) {
+    rgb = overlay->hili_color;
+    for (i = sizeof (overlay->color) / sizeof (overlay->color[0]); i > 0; i--) {
+      clut_t *yuv = (clut_t *)rgb;
+      *rgb++ = frame->yuv2rgb->yuv2rgb_single_pixel_fun (frame->yuv2rgb, yuv->y, yuv->cb, yuv->cr);
     }
     overlay->hili_rgb_clut++;
   }
@@ -451,8 +437,8 @@ static void fb_overlay_clut_yuv2rgb(fb_driver_t *this,
 static void fb_overlay_blend (vo_driver_t *this_gen, vo_frame_t *frame_gen,
 			      vo_overlay_t *overlay)
 {
-  fb_driver_t *this = (fb_driver_t *)this_gen;
-  fb_frame_t *frame = (fb_frame_t *)frame_gen;
+  fb_driver_t *this  = xine_container_of(this_gen,  fb_driver_t, vo_driver);
+  fb_frame_t  *frame = xine_container_of(frame_gen, fb_frame_t,  vo_frame);
 
   this->alphablend_extra_data.offset_x = frame_gen->overlay_offset_x;
   this->alphablend_extra_data.offset_y = frame_gen->overlay_offset_y;
@@ -500,13 +486,14 @@ static void fb_overlay_blend (vo_driver_t *this_gen, vo_frame_t *frame_gen,
 
 static int fb_redraw_needed(vo_driver_t *this_gen)
 {
+  (void)this_gen;
   return 0;
 }
 
 static void fb_display_frame(vo_driver_t *this_gen, vo_frame_t *frame_gen)
 {
-  fb_driver_t  *this = (fb_driver_t *) this_gen;
-  fb_frame_t   *frame = (fb_frame_t *) frame_gen;
+  fb_driver_t  *this  = xine_container_of(this_gen,  fb_driver_t, vo_driver);
+  fb_frame_t   *frame = xine_container_of(frame_gen, fb_frame_t,  vo_frame);
   uint8_t	*dst, *src;
   int y;
 
@@ -563,7 +550,7 @@ static void fb_display_frame(vo_driver_t *this_gen, vo_frame_t *frame_gen)
 
 static int fb_get_property(vo_driver_t *this_gen, int property)
 {
-  fb_driver_t *this = (fb_driver_t *)this_gen;
+  fb_driver_t *this = xine_container_of(this_gen, fb_driver_t, vo_driver);
 
   switch(property)
   {
@@ -605,10 +592,27 @@ static int fb_get_property(vo_driver_t *this_gen, int property)
 
 static int fb_set_property(vo_driver_t *this_gen, int property, int value)
 {
-  fb_driver_t *this = (fb_driver_t *)this_gen;
+  fb_driver_t *this = xine_container_of(this_gen, fb_driver_t, vo_driver);
 
   switch(property)
   {
+    case VO_PROP_DISCARD_FRAMES:
+      if (value == -1) {
+        int n = 0;
+        if (this->old_frame) {
+          this->old_frame->vo_frame.free (&this->old_frame->vo_frame);
+          this->old_frame = NULL;
+          n++;
+        }
+        if (this->cur_frame) {
+          this->cur_frame->vo_frame.free (&this->cur_frame->vo_frame);
+          this->cur_frame = NULL;
+          n++;
+        }
+        value = n;
+      }
+      break;
+
     case VO_PROP_ASPECT_RATIO:
       if(value>=XINE_VO_ASPECT_NUM_RATIOS)
 	value = XINE_VO_ASPECT_AUTO;
@@ -622,7 +626,8 @@ static int fb_set_property(vo_driver_t *this_gen, int property, int value)
       this->yuv2rgb_factory->set_csc_levels (this->yuv2rgb_factory,
                                              this->yuv2rgb_brightness,
                                              this->yuv2rgb_contrast,
-                                             this->yuv2rgb_saturation);
+                                             this->yuv2rgb_saturation,
+                                             CM_DEFAULT);
       break;
 
     case VO_PROP_CONTRAST:
@@ -630,7 +635,8 @@ static int fb_set_property(vo_driver_t *this_gen, int property, int value)
       this->yuv2rgb_factory->set_csc_levels (this->yuv2rgb_factory,
                                              this->yuv2rgb_brightness,
                                              this->yuv2rgb_contrast,
-                                             this->yuv2rgb_saturation);
+                                             this->yuv2rgb_saturation,
+                                             CM_DEFAULT);
       break;
 
     case VO_PROP_SATURATION:
@@ -638,7 +644,8 @@ static int fb_set_property(vo_driver_t *this_gen, int property, int value)
       this->yuv2rgb_factory->set_csc_levels (this->yuv2rgb_factory,
                                              this->yuv2rgb_brightness,
                                              this->yuv2rgb_contrast,
-                                             this->yuv2rgb_saturation);
+                                             this->yuv2rgb_saturation,
+                                             CM_DEFAULT);
       break;
 
     default:
@@ -654,6 +661,7 @@ static void fb_get_property_min_max(vo_driver_t *this_gen,
 {
   /* fb_driver_t *this = (fb_driver_t *) this_gen;  */
 
+  (void)this_gen;
   switch (property) {
   case VO_PROP_BRIGHTNESS:
     *min = -128;
@@ -674,17 +682,27 @@ static void fb_get_property_min_max(vo_driver_t *this_gen,
 static int fb_gui_data_exchange(vo_driver_t *this_gen,
 				int data_type, void *data)
 {
+  (void)this_gen;
+  (void)data_type;
+  (void)data;
   return 0;
 }
 
 static void fb_dispose(vo_driver_t *this_gen)
 {
-  fb_driver_t *this = (fb_driver_t *)this_gen;
+  fb_driver_t *this = xine_container_of(this_gen, fb_driver_t, vo_driver);
 
-  munmap(0, this->mem_size);
-  close(this->fd);
+  if (this->video_mem_base != MAP_FAILED)
+    munmap(this->video_mem_base, this->mem_size);
+
+  if (this->fd >= 0)
+    close(this->fd);
 
   _x_alphablend_free(&this->alphablend_extra_data);
+  _x_vo_scale_cleanup (&this->sc, this->xine->config);
+
+  if (this->yuv2rgb_factory)
+    this->yuv2rgb_factory->dispose (this->yuv2rgb_factory);
 
   free(this);
 }
@@ -837,6 +855,7 @@ static int mode_visual(fb_driver_t *this, config_values_t *config,
 		       struct fb_var_screeninfo *var,
 		       struct fb_fix_screeninfo *fix)
 {
+  (void)config;
   switch(fix->visual)
   {
     case FB_VISUAL_TRUECOLOR:
@@ -891,10 +910,13 @@ static int setup_yuv2rgb(fb_driver_t *this, config_values_t *config,
   this->yuv2rgb_factory = yuv2rgb_factory_init(this->yuv2rgb_mode,
 					       this->yuv2rgb_swap,
 					       this->yuv2rgb_cmap);
+  if (!this->yuv2rgb_factory)
+    return 0;
   this->yuv2rgb_factory->set_csc_levels (this->yuv2rgb_factory,
                                          this->yuv2rgb_brightness,
                                          this->yuv2rgb_contrast,
-                                         this->yuv2rgb_saturation);
+                                         this->yuv2rgb_saturation,
+                                         CM_DEFAULT);
 
   return 1;
 }
@@ -964,33 +986,42 @@ static vo_driver_t *fb_open_plugin(video_driver_class_t *class_gen,
   config_values_t *config;
   fb_driver_t *this;
   fb_class_t *class;
-  fb_visual_t *visual = NULL;
-
-  if (visual_gen) {
-    visual = (fb_visual_t *) visual_gen;
-  }
+  const fb_visual_t *visual = (const fb_visual_t *)visual_gen;
 
   class = (fb_class_t *)class_gen;
-  config = class->config;
+  config = class->xine->config;
 
   /* allocate plugin struct */
   this = calloc(1, sizeof(fb_driver_t));
   if(!this)
     return NULL;
 
+  this->video_mem_base = MAP_FAILED;
+
   _x_alphablend_init(&this->alphablend_extra_data, class->xine);
 
+  _x_vo_scale_init(&this->sc, 0, 0, config);
+
   register_callbacks(this);
 
   this->fd = open_fb_device(config, class->xine);
   if(this->fd == -1)
     goto error;
-  if(!get_fb_var_screeninfo(this->fd, &this->fb_var, class->xine))
+  if(!get_fb_var_screeninfo(this->fd, &this->fb_var, class->xine)) {
+    xprintf(this->xine, XINE_VERBOSITY_LOG, "video_out_fb: get_fb_var_screeninfo() failed: %s\n", strerror(errno));
     goto error;
-  if(!get_fb_fix_screeninfo(this->fd, &this->fb_fix, class->xine))
-    goto error;
-  if (!set_fb_palette (this->fd, &this->fb_var))
+  }
+  if(!get_fb_fix_screeninfo(this->fd, &this->fb_fix, class->xine)) {
+    xprintf(this->xine, XINE_VERBOSITY_LOG, "video_out_fb: get_fb_fix_screeninfo() failed: %s\n", strerror(errno));
     goto error;
+  }
+
+  if (!set_fb_palette (this->fd, &this->fb_var)) {
+    xprintf(this->xine, XINE_VERBOSITY_LOG, "video_out_fb: set_fb_palette() failed: %s\n", strerror(errno));
+    if (this->fb_var.bits_per_pixel < 24) {
+      goto error;
+    }
+  }
 
   this->xine = class->xine;
 
@@ -1001,7 +1032,6 @@ static vo_driver_t *fb_open_plugin(video_driver_class_t *class_gen,
       (this->fb_var.xres_virtual *
        this->fb_var.bits_per_pixel)/8;
 
-  _x_vo_scale_init(&this->sc, 0, 0, config);
   this->sc.gui_width  = this->fb_var.xres;
   this->sc.gui_height = this->fb_var.yres;
   this->sc.user_ratio = XINE_VO_ASPECT_AUTO;
@@ -1033,38 +1063,48 @@ static vo_driver_t *fb_open_plugin(video_driver_class_t *class_gen,
   /* mmap whole video memory */
   this->mem_size = this->fb_fix.smem_len;
   this->video_mem_base = mmap(0, this->mem_size, PROT_READ | PROT_WRITE,
-			      MAP_SHARED, this->fd, 0);
+                              MAP_SHARED, this->fd, 0);
+  if (this->video_mem_base == MAP_FAILED)
+    goto error;
+
   return &this->vo_driver;
+
 error:
-  free(this);
+  xprintf(class->xine, XINE_VERBOSITY_DEBUG,
+          "video_out_fb: Unable to configure fb device, aborting: %s\n", strerror(errno));
+  fb_dispose(&this->vo_driver);
   return 0;
 }
 
-static void *fb_init_class(xine_t *xine, void *visual_gen)
+static void *fb_init_class(xine_t *xine, const void *visual_gen)
 {
-  fb_class_t *this = calloc(1, sizeof(fb_class_t));
+  fb_class_t *this;
+
+  (void)visual_gen;
+
+  this = calloc(1, sizeof(*this));
+  if (!this)
+    return NULL;
 
   this->driver_class.open_plugin     = fb_open_plugin;
   this->driver_class.identifier      = "fb";
   this->driver_class.description     = N_("Xine video output plugin using the Linux frame buffer device");
   this->driver_class.dispose         = default_video_driver_class_dispose;
 
-  this->config          = xine->config;
   this->xine            = xine;
 
   return this;
 }
 
-static const vo_info_t vo_info_fb =
-{
-  1,                    /* priority    */
-  XINE_VISUAL_TYPE_FB   /* visual type */
+static const vo_info_t vo_info_fb = {
+  .priority    = 1,
+  .visual_type = XINE_VISUAL_TYPE_FB,
 };
 
 /* exported plugin catalog entry */
 const plugin_info_t xine_plugin_info[] EXPORTED = {
   /* type, API, "name", version, special_info, init_function */
   { PLUGIN_VIDEO_OUT, 22, "fb", XINE_VERSION_CODE, &vo_info_fb, fb_init_class },
-  { PLUGIN_NONE, 0, "", 0, NULL, NULL }
+  { PLUGIN_NONE, 0, NULL, 0, NULL, NULL }
 };
 
