@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2000-2020 the xine project
+ * Copyright (C) 2000-2022 the xine project
  *
  * This file is part of xine, a free video player.
  *
@@ -48,12 +48,45 @@
 #define SINGLE_CHUNK_ENUMS
 
 #include <xine/xineutils.h>
+#include <xine/sorted_array.h>
 #include <xine/xine_internal.h>
 #include "xine_private.h"
 
+#define MAX_SORT_KEY 320
+
+#if defined(WIN32)
+#  define PSEP '\\'
+#else
+#  define PSEP '/'
+#endif
+
 /* FIXME: static data, no expiry ?! */
 static const xine_config_entry_translation_t *config_entry_translation_user = NULL;
 
+typedef struct {
+  cfg_entry_t entry;
+  int *magic;
+  char *internal_key; /** << xine_fast_string_t * */
+} fat_cfg_entry_t;
+
+static void _config_set_fat_entry (fat_cfg_entry_t *entry) {
+  entry->magic = &entry->entry.range_min;
+}
+
+static int _config_is_fat_entry (fat_cfg_entry_t *entry) {
+  return entry->magic == &entry->entry.range_min;
+}
+
+typedef struct {
+  fat_cfg_entry_t entry;
+  char buf[MAX_SORT_KEY + 32];
+} very_fat_cfg_entry_t;
+
+typedef struct {
+  config_values_t config;
+  xine_sarray_t *key_index;
+} fat_config_values_t;
+
 typedef struct {
   xine_config_cb_t callback;
   void            *data;
@@ -456,68 +489,112 @@ static const char *config_xlate_old (const char *s) {
   return NULL;
 }
 
+static int config_section_enum (const uint8_t *s, uint32_t l) {
+  int d;
 
-static int config_section_enum(const char *sect) {
-  static const char * const known_section[] = {
-    "\x03""audio",
-    "\x08""codec",
-    "\x09""decoder",
-    "\x05""dxr3",
-    "\x0c""effects",
-    "\x0d""engine",
-    "\x01""gui",
-    "\x06""input",
-    "\x07""media",
-    "\x0e""misc",
-    "\x0b""post",
-    "\x0a""subtitles",
-    "\x02""ui",
-    "\x04""video"
-  };
-  int b = 0, e = sizeof (known_section) / sizeof (known_section[0]), m = e >> 1;
-  do {
-    int d = strcmp (sect, known_section[m] + 1);
-    if (d == 0)
-      return known_section[m][0];
-    if (d < 0)
-      e = m;
-    else
-      b = m + 1;
-    m = (b + e) >> 1;
-  } while (b != e);
+  switch (l) {
+    case 2:
+      if (!memcmp (s, "ui", 2))
+        return 2;
+      break;
+    case 3:
+      if (!memcmp (s, "gui", 3))
+        return 1;
+      break;
+    case 4:
+      d = memcmp (s, "misc", 4);
+      if (d == 0)
+        return 14;
+      if (d < 0) {
+        if (!memcmp (s, "dxr3", 4))
+          return 5;
+      } else {
+        if (!memcmp (s, "post", 4))
+          return 11;
+      }
+      break;
+    case 5:
+      d = memcmp (s, "input", 5);
+      if (d == 0)
+        return 6;
+      if (d < 0) {
+        if (!memcmp (s, "audio", 5))
+          return 3;
+        if (!memcmp (s, "codec", 5))
+          return 8;
+      } else {
+        if (!memcmp (s, "media", 5))
+          return 7;
+        if (!memcmp (s, "video", 5))
+          return 4;
+      }
+      break;
+    case 6:
+      if (!memcmp (s, "engine", 6))
+        return 13;
+      break;
+    case 7:
+      if (!memcmp (s, "decoder", 7))
+        return 9;
+      if (!memcmp (s, "effects", 7))
+        return 12;
+      break;
+    case 9:
+      if (!memcmp (s, "subtitles", 9))
+        return 10;
+      break;
+    default: ;
+  }
   return 0;
 }
 
-#define MAX_SORT_KEY 320
-static void config_make_sort_key (char *dest, const char *key, int exp_level) {
-  char *q = dest, *e = dest + MAX_SORT_KEY - 7;
-  const char *p;
+static size_t config_make_sort_key (char *dest, const char *key, int exp_level) {
+  uint8_t *q = (uint8_t *)dest, *e = q + MAX_SORT_KEY - 7;
+  const uint8_t *p = (const uint8_t *)key, *b;
   int n;
-  p = key;
+  uint32_t l;
+
   /* section name */
-  while (q < e) {
-    char z = *p;
-    if (z == 0) break;
+  b = p;
+  while (1) {
+    while (*p > '.')
+      p++;
+    if (!*p || (*p == '.'))
+      break;
     p++;
-    if (z == '.') break;
-    *q++ = z;
   }
-  *q = 0;
-  n = config_section_enum (dest);
-  if (n) {
-    q = dest;
+  l = p - b;
+  n = config_section_enum (b, l);
+  if (n > 0) {
     *q++ = n;
+  } else {
+    if (l > (uint32_t)(e - q))
+      l = e - q;
+    memcpy (q, b, l);
+    q += l;
   }
-  *q++ = 0x1f;
-  /* subsection name */
-  while (q < e) {
-    char z = *p;
-    if (z == 0) break;
+  if (*p)
     p++;
-    if (z == '.') break;
-    *q++ = z;
-  }
   *q++ = 0x1f;
+
+  /* subsection name */
+  b = p;
+  while (1) {
+    while (*p > '.')
+      p++;
+    if (!*p || (*p == '.'))
+      break;
+    p++;
+  }
+  l = p - b;
+  if (l > (uint32_t)(e - q))
+    l = e - q;
+  memcpy (q, b, l);
+  q += l;
+  if (*p)
+    p++;
+  *q++ = 0x1f;
+
   /* TJ. original code did sort by section/subsection/exp/name.
    * We can do that here as well but that means inefficient
    * adding (2 passes) and finding (linear scanning).
@@ -539,16 +616,19 @@ static void config_make_sort_key (char *dest, const char *key, int exp_level) {
 #else
   (void)exp_level;
 #endif
+
   /* entry name */
-  while (q < e) {
-    char z = *p;
-    if (z == 0) break;
-    p++;
-    *q++ = z;
-  }
+  b = p;
+  l = strlen ((const char *)b);
+  if (l > (uint32_t)(e - q))
+    l = e - q;
+  memcpy (q, b, l);
+  q += l;
   *q = 0;
+  return q - (uint8_t *)dest;
 }
 
+#if 0
 /* Ugly: rebuild index every time. Maybe we could cache it somewhere? */
 static cfg_entry_t **config_array (config_values_t *this, cfg_entry_t **tab, int *n) {
   cfg_entry_t *e;
@@ -575,140 +655,127 @@ static cfg_entry_t **config_array (config_values_t *this, cfg_entry_t **tab, int
   *n = i;
   return tab;
 }
-
-#define FIND_ONLY 0x7fffffff
-static cfg_entry_t *config_insert (config_values_t *this, const char *key, int exp_level) {
-  char new_sortkey[MAX_SORT_KEY];
-  char cur_sortkey[MAX_SORT_KEY];
-  cfg_entry_t *entry;
-
-  do {
-    if (!this->first || !this->last) {
-      if (exp_level == FIND_ONLY)
-        return NULL;
-      entry = calloc (1, sizeof (cfg_entry_t));
-      if (!entry)
-        return NULL;
-#ifdef DEBUG_CONFIG_FIND
-      printf ("config_insert (\"%s\", %d) = new (0).\n", key, exp_level);
 #endif
-      this->first = entry;
-      this->last  = entry;
-#ifndef HAVE_ZERO_SAFE_MEM
-      entry->next = NULL;
-#endif
-      break;
-    }
 
-    config_make_sort_key (new_sortkey, key, exp_level);
-
-    /* Most frequent case is loading a config file entry.
-     * Unless edited by user, these come in already sorted.
-     * Thus try last pos first.
-     */
-    if (exp_level != FIND_ONLY) {
-      config_make_sort_key (cur_sortkey, this->last->key, this->last->exp_level);
-      if (strcmp (new_sortkey, cur_sortkey) > 0) {
-        entry = calloc (1, sizeof (cfg_entry_t));
-        if (!entry)
-          return NULL;
-#ifdef DEBUG_CONFIG_FIND
-        printf ("config_insert (\"%s\", %d) = new (last).\n", key, exp_level);
-#endif
-        this->last->next = entry;
-        this->last       = entry;
-#ifndef HAVE_ZERO_SAFE_MEM
-        entry->next      = NULL;
-#endif
-        break;
-      }
-    }
-
-    {
-      cfg_entry_t *aentr[1024], **entries;
-      int n, b, m, e, d;
-      n = 1024;
-      entries = config_array (this, aentr, &n);
-      if (!entries)
-        return NULL;
-
-      b = 0; e = n; m = n >> 1;
-      do {
-        config_make_sort_key (cur_sortkey, entries[m]->key, entries[m]->exp_level);
-        d = strcmp (new_sortkey, cur_sortkey);
-        if (d == 0)
-          break;
-        if (d < 0)
-          e = m;
-        else
-          b = m + 1;
-        m = (b + e) >> 1;
-      } while (b != e);
-
-      if (d == 0) {
-        entry = entries[m];
-        if (entries != aentr)
-          free (entries);
-#ifdef DEBUG_CONFIG_FIND
-        printf ("config_insert (\"%s\", %d) = found (%d/%d).\n", key, exp_level, m, n);
-#endif
-        return entry;
-      }
-
-      if (exp_level == FIND_ONLY) {
-        if (entries != aentr)
-          free (entries);
-#ifdef DEBUG_CONFIG_FIND
-        printf ("config_insert (\"%s\", %d) = not found.\n", key, exp_level);
-#endif
-        return NULL;
-      }
-
-      entry = calloc (1, sizeof (cfg_entry_t));
-      if (!entry) {
-        if (entries != aentr)
-          free (entries);
-        return NULL;
-      }
-      if (m == 0) {
-        entry->next = this->first;
-        this->first = entry;
-      } else {
-        entries[m - 1]->next = entry;
-        if (m < n) {
-          entry->next = entries[m];
-        } else {
-          entry->next = NULL;
-          this->last = entry;
+static int config_validate (config_values_t *this_gen) {
+  fat_config_values_t *this = (fat_config_values_t *)this_gen;
+  fat_cfg_entry_t *entry;
+  int num_new = 0;
+  /* Paranoia:
+   * 1. Find manually added entries. */
+  for (entry = (fat_cfg_entry_t *)this->config.first; entry; entry = (fat_cfg_entry_t *)entry->entry.next) {
+    if (!_config_is_fat_entry (entry)) {
+      if (xine_sarray_add (this->key_index, entry) >= 0) {
+        num_new++;
+        if (this->config.xine) {
+          xprintf (this->config.xine, XINE_VERBOSITY_DEBUG,
+            LOG_MODULE ": WARNING: found manually added entry \"%s\".\n", entry->entry.key);
         }
       }
-      if (entries != aentr)
-        free (entries);
-#ifdef DEBUG_CONFIG_FIND
-      printf ("config_insert (\"%s\", %d) = new (%d/%d).\n", key, exp_level, m, n);
-#endif
     }
-  } while (0);
+  }
+  if (num_new) {
+    int index, num_entries = xine_sarray_size (this->key_index);
+    cfg_entry_t **prev;
 
-#ifndef HAVE_ZERO_SAFE_MEM
-  entry->num_value     = 0;
-  entry->num_default   = 0;
-  entry->range_min     = 0;
-  entry->range_max     = 0;
-  entry->enum_values   = NULL;
-  entry->unknown_value = NULL;
-  entry->str_value     = NULL;
-  entry->str_default   = NULL;
-  entry->help          = NULL;
-  entry->description   = NULL;
-  entry->callback      = NULL;
-  entry->callback_data = NULL;
+    for (prev = &this->config.first, index = 0; index < num_entries; prev = &entry->entry.next, index++) {
+      entry = xine_sarray_get (this->key_index, index);
+
+      *prev = &entry->entry;
+    }
+    this->config.last = &entry->entry;
+  }
+  return num_new;
+}
+
+#define FIND_ONLY 0x7fffffff
+static cfg_entry_t *config_insert (config_values_t *this_gen, const char *key, int exp_level) {
+  fat_config_values_t *this = (fat_config_values_t *)this_gen;
+  very_fat_cfg_entry_t dummy_entry;
+  size_t internal_key_len;
+  fat_cfg_entry_t *entry;
+  int index, num_entries;
+
+  _config_set_fat_entry (&dummy_entry.entry);
+  dummy_entry.entry.internal_key = xine_fast_string_init (dummy_entry.buf, sizeof (dummy_entry.buf));
+  internal_key_len = config_make_sort_key (dummy_entry.entry.internal_key, key, exp_level);
+  xine_fast_string_set (dummy_entry.entry.internal_key, NULL, internal_key_len);
+  num_entries = xine_sarray_size (this->key_index);
+
+  if (exp_level == FIND_ONLY) {
+    index = xine_sarray_binary_search (this->key_index, &dummy_entry);
+    if (index < 0) {
+      /* scan manually added entries */
+      if (config_validate (&this->config))
+        index = xine_sarray_binary_search (this->key_index, &dummy_entry);
+      if (index < 0)
+        return NULL;
+    }
+    entry = xine_sarray_get (this->key_index, index);
+  } else {
+    index = xine_sarray_add (this->key_index, &dummy_entry);
+    if (index >= 0) {
+      cfg_entry_t *e1, *e2;
+      char *buf;
+      /* new */
+      entry = malloc (sizeof (*entry) + internal_key_len + 32);
+      xine_sarray_move_location (this->key_index, entry, index);
+      if (!entry)
+        return NULL;
+#ifdef HAVE_ZERO_SAFE_MEM
+      memset (&entry->entry, 0, sizeof (entry->entry));
+#else
+      entry->entry.num_value     = 0;
+      entry->entry.num_default   = 0;
+      entry->entry.range_min     = 0;
+      entry->entry.range_max     = 0;
+      entry->entry.next          = NULL;
+      entry->entry.enum_values   = NULL;
+      entry->entry.unknown_value = NULL;
+      entry->entry.str_value     = NULL;
+      entry->entry.str_default   = NULL;
+      entry->entry.help          = NULL;
+      entry->entry.description   = NULL;
+      entry->entry.callback      = NULL;
+      entry->entry.callback_data = NULL;
 #endif
-  entry->config        = this;
-  entry->key           = strdup(key);
-  entry->type          = XINE_CONFIG_TYPE_UNKNOWN;
-  entry->exp_level     = exp_level;
-  return entry;
+      entry->entry.config        = &this->config;
+      entry->entry.key           = strdup (key);
+      entry->entry.type          = XINE_CONFIG_TYPE_UNKNOWN;
+      entry->entry.exp_level     = exp_level;
+      _config_set_fat_entry (entry);
+      buf = (char *)entry + sizeof (*entry);
+      entry->internal_key = xine_fast_string_init (buf, internal_key_len + 32);
+      xine_fast_string_set (entry->internal_key, dummy_entry.entry.internal_key, internal_key_len);
+      /* sigh. make public links. */
+      num_entries++;
+      e1 = index > 0 ? xine_sarray_get (this->key_index, index - 1) : NULL;
+      e2 = index < num_entries - 1 ? xine_sarray_get (this->key_index, index + 1) : NULL;
+      if (e1) {
+        if (e2) {
+          if (e1->next == e2) {
+            e1->next = &entry->entry;
+            entry->entry.next = e2;
+          }
+        } else {
+          e1->next = &entry->entry;
+          this->config.last = &entry->entry;
+        }
+      } else {
+        if (e2) {
+          this->config.first = &entry->entry;
+          entry->entry.next = e2;
+        } else {
+          this->config.first = this->config.last = &entry->entry;
+        }
+      }
+    } else {
+      /* found */
+      index = ~index;
+      entry = xine_sarray_get (this->key_index, index);
+    }
+  }
+  return &entry->entry;
 }
 
 static const char *config_xlate_internal (const char *key, const xine_config_entry_translation_t *trans)
@@ -720,22 +787,21 @@ static const char *config_xlate_internal (const char *key, const xine_config_ent
   return NULL;
 }
 
-static const char *config_translate_key (const char *key, char **tmp) {
+#define _MAX_CFG_KEY 320
+static const char *config_translate_key (const char *key, char *tmp, size_t klen) {
   /* Returns translated key or, if no translation found, NULL.
-   * Translated key may be in a static buffer allocated within this function.
-   * NOT re-entrant; assumes that config_lock is held.
-   */
-  unsigned trans;
-  const char *newkey = NULL;
+   * Translated key may be written to tmp. */
+  const char *newkey;
 
   /* first, special-case the decoder entries (so that new ones can be added
-   * without requiring modification of the translation table)
-   */
-  *tmp = NULL;
-  if (!strncmp (key, "decoder.", 8) &&
-      !strcmp (key + (trans = strlen (key)) - 9, "_priority")) {
-    *tmp = _x_asprintf ("engine.decoder_priorities.%.*s", trans - 17, key + 8);
-    return *tmp;
+   * without requiring modification of the translation table). */
+  if (!strncmp (key, "decoder.", 8)) {
+    if ((klen > 8 + 9) && !memcmp (key + klen - 9, "_priority", 9) && (klen < _MAX_CFG_KEY + 8 + 9 - 26 - 1)) {
+      memcpy (tmp, "engine.decoder_priorities.", 26);
+      memcpy (tmp + 26, key + 8, klen - 8 - 9);
+      tmp[26 + klen - 8 - 9] = 0;
+      return tmp;
+    }
   }
 
   /* search the translation table... */
@@ -748,7 +814,7 @@ static const char *config_translate_key (const char *key, char **tmp) {
 
 static cfg_entry_t *config_lookup_entry_int (config_values_t *this, const char *key) {
   cfg_entry_t *entry;
-  char *tmp = NULL;
+  char tmp[_MAX_CFG_KEY];
 
   /* try twice at most (second time with translation from old key name) */
   entry = config_insert (this, key, FIND_ONLY);
@@ -756,13 +822,10 @@ static cfg_entry_t *config_lookup_entry_int (config_values_t *this, const char *
     return entry;
   /* we did not find a match, maybe this is an old config entry name
    * trying to translate */
-  key = config_translate_key (key, &tmp);
-  if (!key) {
-    free (tmp);
+  key = config_translate_key (key, tmp, strlen (key));
+  if (!key)
     return NULL;
-  }
   entry = config_insert (this, key, FIND_ONLY);
-  free (tmp);
   return entry;
 }
 
@@ -1444,64 +1507,36 @@ void xine_config_set_translation_user (const xine_config_entry_translation_t *xl
  * load/save config data from/to afile (e.g. $HOME/.xine/config)
  */
 void xine_config_load (xine_t *xine, const char *filename) {
-
   config_values_t *this = xine->config;
+  xine_fast_text_t *xft;
+
   this->xine = xine;
 
   lprintf ("reading from file '%s'\n", filename);
 
-  do {
-    FILE *f_config;
+  /* TJ. I got far less than 32k, so > 2M is probably insane. */
+  xft = xine_fast_text_load (filename, 2 << 20);
+  if (xft) {
     int version;
-    char *buf, *line, *nextline;
-
-    f_config = fopen (filename, "rb");
-    if (!f_config)
-      break;
-    {
-      long int flen;
-      fseek (f_config, 0, SEEK_END);
-      flen = ftell (f_config);
-      if (flen < 0) {
-        fclose (f_config);
-        break;
-      }
-      /* TJ. I got far less than 32k, so > 2M is probably insane. */
-      if (flen > (2 << 20))
-        flen = 2 << 20;
-      buf = malloc (flen + 2);
-      if (!buf) {
-        fclose (f_config);
-        break;
-      }
-      fseek (f_config, 0, SEEK_SET);
-      flen = fread (buf, 1, flen, f_config);
-      buf[flen] = '\n';
-      buf[flen + 1] = 0;
-    }
-    fclose (f_config);
 
     pthread_mutex_lock (&this->config_lock);
     version = this->current_version;
     pthread_mutex_unlock (&this->config_lock);
 
-    for (line = buf; line[0]; line = nextline) {
+    while (1) {
+      size_t lsize;
+      char *line = xine_fast_text_line (xft, &lsize);
       char *value;
-      /* make string from line */
-      char *lend = strchr (line, '\n');
-      if (!lend) /* should not happen */
+
+      if (!line)
         break;
-      nextline = lend + 1;
-      if ((lend > line) && (lend[-1] == '\r'))
-        lend--;
-      lend[0] = 0;
 
       /* skip comments */
       if (line[0] == '#')
         continue;
 
       if (line[0] == '.') {
-        if (!strncmp (line, ".version:", 9)) {
+        if ((lsize > 9) && !memcmp (line + 1, "version:", 8)) {
           const char *val = line + 9;
           version = xine_str2int32 (&val);
           if (version > CONFIG_FILE_VERSION) {
@@ -1518,6 +1553,7 @@ void xine_config_load (xine_t *xine, const char *filename) {
       value = strchr (line, ':');
       if (value) {
         cfg_entry_t *entry;
+        size_t klen = value - line;
 
         *value++ = 0;
 
@@ -1527,12 +1563,11 @@ void xine_config_load (xine_t *xine, const char *filename) {
           /* old config file -> let's see if we have to rename this one */
           entry = config_insert (this, line, FIND_ONLY);
           if (!entry) {
-            char *tmp = NULL;
-            const char *key = config_translate_key (line, &tmp);
+            char tmp[_MAX_CFG_KEY];
+            const char *key = config_translate_key (line, tmp, klen);
             if (!key)
               key = line; /* no translation? fall back on untranslated key */
             entry = config_insert (this, key, 50);
-            free (tmp);
           }
         } else {
           entry = config_insert (this, line, 50);
@@ -1544,85 +1579,133 @@ void xine_config_load (xine_t *xine, const char *filename) {
         pthread_mutex_unlock (&this->config_lock);
       }
     }
-    free (buf);
+    xine_fast_text_unload (&xft);
     xine_log (xine, XINE_LOG_MSG,
       _("Loaded configuration from file '%s'\n"), filename);
     return;
-  } while (0);
+  }
 
   if (errno != ENOENT)
     xine_log (xine, XINE_LOG_MSG,
       _("Failed to load configuration from file '%s': %s\n"), filename, strerror (errno));
 }
 
-void xine_config_save (xine_t *xine, const char *filename) {
+static size_t xine_realpath (char *buf, const char *filename, size_t bsize, int *num_links) {
+  char tbuf[1536];
+  struct stat sbuf;
+  size_t used;
+  int try;
 
-  config_values_t *this = xine->config;
-  char             temp[XINE_PATH_MAX];
-  int              backup = 0;
-  struct stat      backup_stat, config_stat;
-  FILE            *f_config, *f_backup;
-
-  snprintf(temp, XINE_PATH_MAX, "%s~", filename);
-  unlink (temp);
-
-  if (stat(temp, &backup_stat) != 0) {
-
-    lprintf("backing up configfile to %s\n", temp);
-
-    f_backup = fopen(temp, "wb");
-    f_config = fopen(filename, "rb");
-
-    if (f_config && f_backup && (stat(filename, &config_stat) == 0)) {
-      char    *buf = NULL;
-      size_t   rlen;
-
-      buf = (char *) malloc(config_stat.st_size + 1);
-      if((rlen = fread(buf, 1, config_stat.st_size, f_config)) && ((off_t)rlen == config_stat.st_size)) {
-	if (rlen != fwrite(buf, 1, rlen, f_backup)) {
-	  lprintf("backing up configfile to %s failed\n", temp);
-	}
-      }
-      free(buf);
-
-      fclose(f_config);
-      fclose(f_backup);
-
-      if (stat(temp, &backup_stat) == 0 && config_stat.st_size == backup_stat.st_size)
-	backup = 1;
-      else
-	unlink(temp);
-
-    }
-    else {
-
-      if (f_config)
-        fclose(f_config);
-      else
-	backup = 1;
-
-      if (f_backup)
-        fclose(f_backup);
-
-    }
+  if (!filename || !buf || !bsize)
+    return 0;
+  if (!filename) {
+    buf[0] = 0;
+    return 0;
   }
 
-  if (!backup && (stat(filename, &config_stat) == 0)) {
-    xprintf(xine, XINE_VERBOSITY_LOG, _("configfile: WARNING: backing up configfile to %s failed\n"), temp);
+  used = strlcpy (buf, filename, bsize);
+  if (used >= bsize)
+    used = bsize - 1;
+
+  try = 8;
+#if defined(HAVE_LSTAT) && defined(HAVE_READLINK)
+  for (; try; try--) {
+    ssize_t r;
+
+    if (lstat (buf, &sbuf))
+      break;
+    if (!S_ISLNK (sbuf.st_mode))
+      break;
+    r = readlink (buf, tbuf, sizeof (tbuf) - 1);
+    if (r <= 0)
+      break;
+    tbuf[r] = 0;
+    if (tbuf[0] == PSEP) {
+      /* absolute link */
+      used = (size_t)r < bsize ? (size_t)r : bsize - 1;
+      memcpy (buf, tbuf, used);
+      buf[used] = 0;
+    } else {
+      size_t left;
+      char *p;
+      /* relative link */
+      for (p = buf + used; (p > buf) && (p[-1] != PSEP); p--) ;
+      left = bsize - (p - buf);
+      used = (size_t)r < left ? (size_t)r : left - 1;
+      memcpy (p, tbuf, used);
+      p[used] = 0;
+      used += p - buf;
+    }
+  }
+  if (!try)
+    return 0;
+#endif
+  *num_links = 8 - try;
+
+  return used;
+}
+
+void xine_config_save (xine_t *xine, const char *filename) {
+  config_values_t *this;
+  char fname[1536], bname[1536], tname[1536];
+  struct stat sbuf;
+  size_t blen;
+  FILE *f;
+#define XCF_HAVE_BACKUP 1
+#define XCF_HAVE_FILE   2
+#define XCF_HAVE_TFILE  4
+#define XCF_HAVE_ITEMS  8
+  uint32_t flags = 0;
+  int num_links = 0;
+
+  if (!xine || !filename)
+    return;
+
+  /* yes this _is_ relevant. */
+  blen = xine_realpath (fname, filename, sizeof (fname) - 48, &num_links);
+  if (!blen)
+    return;
+  if (num_links) {
+    xprintf (xine, XINE_VERBOSITY_DEBUG,
+      LOG_MODULE ": %s -> %s.\n", filename, fname);
+  }
+
+  /* When X server shuts down while multiple xine instances run,
+   * a) concurrent writes may trash the file, and/or
+   * b) write may be interrupted. */
+
+  this = xine->config;
+  xine_fast_memcpy (bname, fname, blen);
+  memcpy (bname + blen, "~", 2);
+  xine_fast_memcpy (tname, fname, blen);
+  {
+    char *p = tname + blen;
+
+    *p++ = '.';
+    xine_uint32_2str (&p, (uintptr_t)getpid ());
+    *p++ = '.';
+    xine_uint32_2str (&p, (uintptr_t)this);
+    *p = 0;
+  }
+
+/*  xprintf(xine, XINE_VERBOSITY_LOG, _("configfile: WARNING: backing up configfile to %s failed\n"), bname); */
+  xprintf (xine, XINE_VERBOSITY_DEBUG, LOG_MODULE ": writing %s.\n", tname);
+
+  f = fopen (tname, "wb");
+  if (!f) {
+    int e = errno;
+
+    xprintf (xine, XINE_VERBOSITY_LOG,
+      LOG_MODULE ": %s: %s (%d).\n", tname, strerror (e), e);
     xprintf(xine, XINE_VERBOSITY_LOG, _("configfile: WARNING: your configuration will not be saved\n"));
     return;
-  }
-
-  lprintf ("writing config file to %s\n", filename);
-
-  f_config = fopen (filename, "wb");
-
-  if (f_config) {
-
+  } else {
 #define XCS_BUF_SIZE 4096
     char buf[XCS_BUF_SIZE], *q, *e = buf + XCS_BUF_SIZE - 28 - 4 * XINE_MAX_INT32_STR;
     cfg_entry_t *entry;
+    size_t flen = 0;
 
+    flags |= XCF_HAVE_TFILE;
     q = buf;
     memcpy (q, "#\n# xine config file\n#\n.version:", 32); q += 32;
     xine_uint32_2str (&q, CONFIG_FILE_VERSION);
@@ -1633,7 +1716,7 @@ void xine_config_save (xine_t *xine, const char *filename) {
       "# Remove the \'#\' at the beginning of the line, if you want to change them.\n"
       "\n",
       151); q += 151;
-    fwrite (buf, 1, q - buf, f_config);
+    flen += fwrite (buf, 1, q - buf, f);
 
     pthread_mutex_lock(&this->config_lock);
 
@@ -1756,32 +1839,51 @@ void xine_config_save (xine_t *xine, const char *filename) {
         memcpy (q, "# bool", 6); q += 6;
         goto tail_num;
       }
-      fwrite (buf, 1, q - buf, f_config);
-
+      flen += fwrite (buf, 1, q - buf, f);
+      flags |= XCF_HAVE_ITEMS;
     }
     pthread_mutex_unlock(&this->config_lock);
 
-    if (fclose(f_config) != 0) {
-      xprintf(xine, XINE_VERBOSITY_LOG, _("configfile: WARNING: writing configuration to %s failed\n"), filename);
-      xprintf(xine, XINE_VERBOSITY_LOG, _("configfile: WARNING: removing possibly broken config file %s\n"), filename);
-      xprintf(xine, XINE_VERBOSITY_LOG, _("configfile: WARNING: you should check the backup file %s\n"), temp);
-      /* writing config failed -> remove file, it might be broken ... */
-      unlink(filename);
-      /* ... but keep the backup */
-      backup = 0;
+    /* paranoia ... */
+    if (fclose (f))
+      flags &= ~XCF_HAVE_TFILE;
+    if (!stat (tname, &sbuf)) {
+      if ((off_t)flen != sbuf.st_size)
+        flags &= ~XCF_HAVE_TFILE;
+    } else {
+      flags &= ~XCF_HAVE_TFILE;
     }
-  }
 
-  if (backup)
-    unlink(temp);
+    if ((flags & (XCF_HAVE_TFILE | XCF_HAVE_ITEMS)) != (XCF_HAVE_TFILE | XCF_HAVE_ITEMS)) {
+      xprintf (xine, XINE_VERBOSITY_LOG, _("configfile: WARNING: writing configuration to %s failed\n"), fname);
+      xprintf (xine, XINE_VERBOSITY_LOG, _("configfile: WARNING: removing possibly broken config file %s\n"), tname);
+      xprintf (xine, XINE_VERBOSITY_LOG, _("configfile: WARNING: you should check the backup file %s\n"), bname);
+      /* writing config failed -> remove file, it might be broken. */
+      unlink (tname);
+      return;
+    }
+
+    if (!stat (bname, &sbuf) && S_ISREG (sbuf.st_mode))
+      flags |= XCF_HAVE_BACKUP;
+    if (!stat (fname, &sbuf) && S_ISREG (sbuf.st_mode))
+      flags |= XCF_HAVE_FILE;
+
+    if (flags & XCF_HAVE_FILE) {
+      if (flags & XCF_HAVE_BACKUP)
+        unlink (bname);
+      rename (fname, bname);
+    }
+    rename (tname, fname);
+  }
 }
 
-static void config_dispose (config_values_t *this) {
+static void config_dispose (config_values_t *this_gen) {
+  fat_config_values_t *this = (fat_config_values_t *)this_gen;
   cfg_entry_t *entry, *last;
   int n;
 
-  pthread_mutex_lock (&this->config_lock);
-  entry = this->first;
+  pthread_mutex_lock (&this->config.config_lock);
+  entry = this->config.first;
 
   lprintf ("dispose\n");
 
@@ -1791,7 +1893,7 @@ static void config_dispose (config_values_t *this) {
     entry = entry->next;
 
     last->next = NULL;
-    n += _cfg_cb_clear_report (this->xine, last);
+    n += _cfg_cb_clear_report (this->config.xine, last);
     _x_freep (&last->key);
     _x_freep (&last->unknown_value);
 
@@ -1800,13 +1902,16 @@ static void config_dispose (config_values_t *this) {
     free (last);
   }
 
-  pthread_mutex_unlock (&this->config_lock);
+  xine_sarray_delete (this->key_index);
+  this->key_index = NULL;
 
-  if (n && this->xine) {
-    xprintf (this->xine, XINE_VERBOSITY_DEBUG, "configfile: unregistered %d orphaned change callbacks.\n", n);
+  pthread_mutex_unlock (&this->config.config_lock);
+
+  if (n && this->config.xine) {
+    xprintf (this->config.xine, XINE_VERBOSITY_DEBUG, "configfile: unregistered %d orphaned change callbacks.\n", n);
   }
 
-  pthread_mutex_destroy (&this->config_lock);
+  pthread_mutex_destroy (&this->config.config_lock);
   free (this);
 }
 
@@ -2142,12 +2247,32 @@ static char* config_register_serialized_entry (config_values_t *this, const char
   return NULL;
 }
 
+static int _config_fat_entry_cmp (void *a, void *b) {
+  char buf1[MAX_SORT_KEY + 32], buf2[MAX_SORT_KEY + 32], *fs1, *fs2;
+  fat_cfg_entry_t *d = (fat_cfg_entry_t *)a;
+  fat_cfg_entry_t *e = (fat_cfg_entry_t *)b;
+
+  if (_config_is_fat_entry (d)) {
+    fs1 = d->internal_key;
+  } else {
+    fs1 = xine_fast_string_init (buf1, sizeof (buf1));
+    xine_fast_string_set (fs1, NULL, config_make_sort_key (fs1, d->entry.key, d->entry.exp_level));
+  }
+  if (_config_is_fat_entry (e)) {
+    fs2 = e->internal_key;
+  } else {
+    fs2 = xine_fast_string_init (buf2, sizeof (buf2));
+    xine_fast_string_set (fs2, NULL, config_make_sort_key (fs2, e->entry.key, e->entry.exp_level));
+  }
+  return xine_fast_string_cmp (fs1, fs2);
+}
+
 config_values_t *_x_config_init (void) {
 
 #ifdef HAVE_IRIXAL
   volatile /* is this a (old, 2.91.66) irix gcc bug?!? */
 #endif
-  config_values_t *this;
+  fat_config_values_t *this;
   pthread_mutexattr_t attr;
 
   this = calloc (1, sizeof (*this));
@@ -2157,10 +2282,10 @@ config_values_t *_x_config_init (void) {
   }
 
 #ifndef HAVE_ZERO_SAFE_MEM
-  this->first           = NULL;
-  this->last            = NULL;
-  this->current_version = 0;
-  this->xine            = NULL;
+  this->config.first           = NULL;
+  this->config.last            = NULL;
+  this->config.current_version = 0;
+  this->config.xine            = NULL;
 #endif
 
   /* warning: config_lock is a recursive mutex. it must NOT be
@@ -2168,31 +2293,34 @@ config_values_t *_x_config_init (void) {
    */
   pthread_mutexattr_init(&attr);
   pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE);
-  pthread_mutex_init(&this->config_lock, &attr);
+  pthread_mutex_init (&this->config.config_lock, &attr);
   pthread_mutexattr_destroy(&attr);
 
-  this->register_string           = config_register_string;
-  this->register_filename         = config_register_filename;
-  this->register_range            = config_register_range;
-  this->register_enum             = config_register_enum;
-  this->register_num              = config_register_num;
-  this->register_bool             = config_register_bool;
-  this->register_serialized_entry = config_register_serialized_entry;
-  this->update_num                = config_update_num;
-  this->update_string             = config_update_string;
-  this->parse_enum                = config_parse_enum;
-  this->lookup_entry              = config_lookup_entry;
-  this->unregister_callback       = config_unregister_cb;
-  this->dispose                   = config_dispose;
-  this->set_new_entry_callback    = config_set_new_entry_callback;
-  this->unset_new_entry_callback  = config_unset_new_entry_callback;
-  this->get_serialized_entry      = config_get_serialized_entry;
-  this->unregister_callbacks      = config_unregister_callbacks;
-  this->lookup_num                = config_lookup_num;
-  this->lookup_string             = config_lookup_string;
-  this->free_string               = config_free_string;
+  this->config.register_string           = config_register_string;
+  this->config.register_filename         = config_register_filename;
+  this->config.register_range            = config_register_range;
+  this->config.register_enum             = config_register_enum;
+  this->config.register_num              = config_register_num;
+  this->config.register_bool             = config_register_bool;
+  this->config.register_serialized_entry = config_register_serialized_entry;
+  this->config.update_num                = config_update_num;
+  this->config.update_string             = config_update_string;
+  this->config.parse_enum                = config_parse_enum;
+  this->config.lookup_entry              = config_lookup_entry;
+  this->config.unregister_callback       = config_unregister_cb;
+  this->config.dispose                   = config_dispose;
+  this->config.set_new_entry_callback    = config_set_new_entry_callback;
+  this->config.unset_new_entry_callback  = config_unset_new_entry_callback;
+  this->config.get_serialized_entry      = config_get_serialized_entry;
+  this->config.unregister_callbacks      = config_unregister_callbacks;
+  this->config.lookup_num                = config_lookup_num;
+  this->config.lookup_string             = config_lookup_string;
+  this->config.free_string               = config_free_string;
 
-  return this;
+  this->key_index = xine_sarray_new (1024, _config_fat_entry_cmp);
+  xine_sarray_set_mode (this->key_index, XINE_SARRAY_MODE_UNIQUE);
+
+  return &this->config;
 }
 
 int _x_config_change_opt(config_values_t *config, const char *opt) {
