@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2000-2020 the xine project
+ * Copyright (C) 2000-2021 the xine project
  *
  * This file is part of xine, a free video player.
  *
@@ -360,18 +360,17 @@ static void *demux_loop (void *stream_gen) {
 
   /* do-while needed to seek after demux finished */
   do {
+    uint32_t input_caps = 0;
+
     xine_gettime (&seek_time);
 
-    {
-      uint32_t input_caps = 0;
-      /* tell xine_play_internal () whether we can seek.
-       * so it may flush fifos early, and suspend us faster. */
-      if (stream->s.input_plugin)
-        input_caps = stream->s.input_plugin->get_capabilities (stream->s.input_plugin);
-      pthread_mutex_lock (&stream->demux.action_lock);
-      stream->demux.input_caps = input_caps;
-      pthread_mutex_unlock (&stream->demux.action_lock);
-    }
+    /* tell xine_play_internal () whether we can seek.
+     * so it may flush fifos early, and suspend us faster. */
+    if (stream->s.input_plugin)
+      input_caps = stream->s.input_plugin->get_capabilities (stream->s.input_plugin);
+    pthread_mutex_lock (&stream->demux.action_lock);
+    stream->demux.input_caps = input_caps;
+    pthread_mutex_unlock (&stream->demux.action_lock);
 
     /* main demuxer loop */
     status = stream->demux.plugin->get_status (stream->demux.plugin);
@@ -380,10 +379,24 @@ static void *demux_loop (void *stream_gen) {
       iterations++;
       status = stream->demux.plugin->send_chunk (stream->demux.plugin);
 
+      if (!(iterations & 31)) {
+        uint32_t new_caps = stream->s.input_plugin
+                          ? stream->s.input_plugin->get_capabilities (stream->s.input_plugin)
+                          : 0;
+        if (new_caps != input_caps) {
+          input_caps = new_caps;
+          pthread_mutex_lock (&stream->demux.action_lock);
+          stream->demux.input_caps = input_caps;
+          pthread_mutex_unlock (&stream->demux.action_lock);
+        }
+      }
+
       /* someone may want to interrupt us */
       if (stream->demux.action_pending > 0) {
         pthread_mutex_lock (&stream->demux.action_lock);
         if (stream->demux.action_pending > 0) {
+          if (stream->s.input_plugin)
+            stream->demux.input_caps = stream->s.input_plugin->get_capabilities (stream->s.input_plugin);
           pthread_mutex_unlock (&stream->demux.lock);
           do {
             pthread_cond_wait (&stream->demux.resume, &stream->demux.action_lock);
@@ -551,6 +564,14 @@ static void *demux_loop (void *stream_gen) {
   return NULL;
 }
 
+int _x_demux_called_from (xine_stream_t *s) {
+  xine_stream_private_t *stream = (xine_stream_private_t *)s;
+
+  if (!stream->demux.thread_running)
+    return 0;
+  return pthread_equal (pthread_self (), stream->demux.thread);
+}
+
 int _x_demux_start_thread (xine_stream_t *s) {
 
   xine_stream_private_t *stream = (xine_stream_private_t *)s;
@@ -635,6 +656,11 @@ int _x_demux_read_header (input_plugin_t *input, void *buffer, off_t size) {
 
   caps = input->get_capabilities (input);
 
+  if ((caps & INPUT_CAP_SIZED_PREVIEW) && (want_size >= (int)sizeof (want_size))) {
+    memcpy (buffer, &want_size, sizeof (want_size));
+    return input->get_optional_data (input, buffer, INPUT_OPTIONAL_DATA_SIZED_PREVIEW);
+  }
+
   if (caps & INPUT_CAP_SEEKABLE) {
     if (input->seek (input, 0, SEEK_SET) != 0)
       return 0;
@@ -644,11 +670,6 @@ int _x_demux_read_header (input_plugin_t *input, void *buffer, off_t size) {
     return want_size;
   }
 
-  if ((caps & INPUT_CAP_SIZED_PREVIEW) && (want_size >= (int)sizeof (want_size))) {
-    memcpy (buffer, &want_size, sizeof (want_size));
-    return input->get_optional_data (input, buffer, INPUT_OPTIONAL_DATA_SIZED_PREVIEW);
-  }
-
   if (caps & INPUT_CAP_PREVIEW) {
     if (want_size < MAX_PREVIEW_SIZE) {
       int read_size;
